from .utils import is_test


def test_is_test():
    test_names = {"some-uuid": {"check-mirrors", "check-seatbelt"}}
    column_names = {"some-uuid": {"carthago", "delenda", "est"}}

    happy_node_test = is_test(
        query_result={"name": "check-mirrors"},
        node_id="some-uuid",
        test_names=test_names,
        column_names=column_names,
    )
    assert happy_node_test is True

    unknown_node_test = is_test(
        query_result={"name": "check-engine"},
        node_id="some-uuid",
        test_names=test_names,
        column_names=column_names,
    )
    assert unknown_node_test is False

    happy_column_test_unique = is_test(
        query_result={"name": "carthago: Unique"},
        node_id="some-uuid",
        test_names=test_names,
        column_names=column_names,
    )
    assert happy_column_test_unique is True

    happy_column_test_null = is_test(
        query_result={"name": "carthago: Null"},
        node_id="some-uuid",
        test_names=test_names,
        column_names=column_names,
    )
    assert happy_column_test_null is True

    unknown_column_test = is_test(
        query_result={"name": "rome: Unique"},
        node_id="some-uuid",
        test_names=test_names,
        column_names=column_names,
    )
    assert unknown_column_test is False

    unknown_node_id_test = is_test(
        query_result={"name": "whatever: Unique"},
        node_id="unknown-uuid",
        test_names=test_names,
        column_names=column_names,
    )
    assert unknown_node_id_test is False



from .enums import LineageEntity
from .lineage import LineageProcessor, valid_lineage

_OLDER_DATE = "2025-01-01 00:00:01.0"
_CLOSER_DATE = "2025-01-01 02:02:02.0"

_TABLE_LINEAGES = [
    {
        "source_table_full_name": "a.b.source",
        "target_table_full_name": "a.b.target",
        "event_time": _CLOSER_DATE,
        "other": "more recent stuff",
    },
    {
        "source_table_full_name": "a.b.source",
        "target_table_full_name": "a.b.target",
        "event_time": _OLDER_DATE,
        "other": "stuff that's too old",
    },
    {
        "source_table_full_name": "no target",
        "target_table_full_name": None,
        "event_time": _CLOSER_DATE,
    },
    {
        "source_table_full_name": None,
        "target_table_full_name": "no source",
        "event_time": _CLOSER_DATE,
    },
]


_COLUMN_LINEAGES = [
    {
        "source_table_full_name": "a.b.source",
        "source_column_name": "src_col",
        "target_table_full_name": "a.b.target",
        "target_column_name": "trgt_col",
        "event_time": _OLDER_DATE,
        "other": "old stuff",
    },
    {
        "source_table_full_name": "a.b.source",
        "source_column_name": "src_col",
        "target_table_full_name": "a.b.target",
        "target_column_name": "trgt_col",
        "event_time": _CLOSER_DATE,
        "other": "newer stuff",
    },
    {
        "source_table_full_name": "a.b.toto",
        "source_column_name": "toto_col",
        "target_table_full_name": "a.b.tata",
        "target_column_name": "tata_col",
        "event_time": _OLDER_DATE,
    },
    {
        "source_table_full_name": "a.b.source",
        "source_column_name": "a.b.source",
        "target_table_full_name": None,
        "target_column_name": None,
        "event_time": _CLOSER_DATE,
    },
]


def test_valid_lineage():
    table_links = valid_lineage(_TABLE_LINEAGES, LineageEntity.TABLE)

    assert len(table_links) == 1
    assert table_links[0]["source_table_full_name"] == "a.b.source"
    assert table_links[0]["target_table_full_name"] == "a.b.target"
    assert table_links[0]["event_time"] == _CLOSER_DATE
    assert table_links[0]["other"] == "more recent stuff"


def test_LineageLinks_add():
    deduplicated_lineage = LineageProcessor(LineageEntity.COLUMN)
    for link in _COLUMN_LINEAGES:
        deduplicated_lineage.add(link)

    lineage = deduplicated_lineage.lineage
    assert len(lineage) == 2
    assert ("a.b.source.src_col", "a.b.target.trgt_col") in lineage
    assert ("a.b.toto.toto_col", "a.b.tata.tata_col") in lineage
    assert (
        lineage[("a.b.source.src_col", "a.b.target.trgt_col")]["other"]
        == "newer stuff"
    )



import logging
from http import HTTPStatus
from typing import Iterator, Optional

from requests import ConnectionError

from ....utils import (
    APIClient,
    BearerAuth,
    RequestSafeMode,
    SerializedAsset,
)
from ..assets import CoalesceAsset, CoalesceQualityAsset
from .credentials import CoalesceCredentials
from .endpoint import (
    CoalesceEndpointFactory,
)
from .type import NodeIDToNamesMapping
from .utils import column_names_per_node, is_test, test_names_per_node

_LIMIT_MAX = 1_000
_MAX_ERRORS = 200

logger = logging.getLogger(__name__)


def _run_result_payload(result: dict, query_result: dict) -> dict:
    return {
        "node_id": result["nodeID"],
        "node_name": result["name"],
        "test_name": query_result["name"],
        "start_time": query_result["startTime"],
        "end_time": query_result["endTime"],
        "status": query_result["status"],
        "success": query_result["success"],
        "isRunning": query_result["isRunning"],
    }


COALESCE_SAFE_MODE = RequestSafeMode(
    status_codes=(HTTPStatus.INTERNAL_SERVER_ERROR,),
    max_errors=_MAX_ERRORS,
)
COALESCE_TIMEOUT_SECONDS = 90


class CoalesceBearerAuth(BearerAuth):
    """Bearer Authentication for Coalesce"""

    def fetch_token(self) -> Optional[str]:
        pass

    def __init__(self, token: str):
        self._token = token


class CoalesceClient(APIClient):
    """REST API client to extract data from Coalesce"""

    def __init__(
        self,
        credentials: CoalesceCredentials,
    ):
        auth = CoalesceBearerAuth(token=credentials.token)
        super().__init__(
            host=credentials.host,
            auth=auth,
            safe_mode=COALESCE_SAFE_MODE,
            timeout=COALESCE_TIMEOUT_SECONDS,
        )

    def _fetch_environments(self) -> Iterator[dict]:
        endpoint = CoalesceEndpointFactory.environments()
        result = self._get(endpoint=endpoint)
        return result["data"]

    def _node_details(self, environment_id: int, node_id: str) -> dict:
        endpoint = CoalesceEndpointFactory.nodes(
            environment_id=environment_id, node_id=node_id
        )
        return self._get(endpoint=endpoint)

    def _fetch_env_nodes(self, environment_id: int) -> SerializedAsset:
        endpoint = CoalesceEndpointFactory.nodes(environment_id=environment_id)
        result = self._get(endpoint=endpoint)
        nodes: list[dict] = []
        for node in result["data"]:
            try:
                details = self._node_details(environment_id, node["id"])
                nodes.append({**node, **details})
            except ConnectionError as e:
                node_id = node["id"]
                message = f"ConnectionError, environment: {environment_id}, node: {node_id}"
                logger.warning(message)
                raise e
        return nodes

    def _fetch_all_nodes(self) -> SerializedAsset:
        nodes: list[dict] = []
        for environment in self._fetch_environments():
            environment_id = environment["id"]
            nodes.extend(self._fetch_env_nodes(environment_id))
        return nodes

    def _fetch_runs(self, starting_from: str) -> SerializedAsset:
        """
        fetch runs, per environment;
        we break per environment to lower the chance of exceeding the 1k limit
        """
        runs: list[dict] = []
        for environment in self._fetch_environments():
            environment_id = environment["id"]
            runs.extend(
                self._fetch_recent_runs_per_env(environment_id, starting_from)
            )
        return runs

    def _fetch_recent_runs_per_env(
        self, environment_id: int, starting_from: str
    ) -> SerializedAsset:
        endpoint = CoalesceEndpointFactory.runs()
        params = {
            "environmentID": environment_id,
            "limit": _LIMIT_MAX,
            "orderBy": "runEndTime",
            "orderByDirection": "asc",
            "startingFrom": starting_from,
        }
        result = self._get(endpoint=endpoint, params=params)
        return result["data"]

    def _fetch_run_results(self, run_id: str) -> SerializedAsset:
        endpoint = CoalesceEndpointFactory.run_results(run_id)
        result = self._get(endpoint=endpoint)
        return result["data"]

    def _run_results_by_run(
        self,
        run_id: str,
        test_names: NodeIDToNamesMapping,
        column_names: NodeIDToNamesMapping,
    ) -> SerializedAsset:
        run_results: list[dict] = []
        for result in self._fetch_run_results(run_id):
            node_id = result["nodeID"]
            for query_result in result["queryResults"]:
                _is_test = is_test(
                    query_result,
                    node_id,
                    test_names,
                    column_names,
                )
                if not _is_test:
                    continue
                run_result = _run_result_payload(result, query_result)
                run_results.append(run_result)
        return run_results

    def _run_results_by_env(
        self, environment_id: int, starting_from: str
    ) -> SerializedAsset:
        run_results: list[dict] = []
        nodes = self._fetch_env_nodes(environment_id)
        test_names = test_names_per_node(nodes)
        column_names = column_names_per_node(nodes)
        runs = self._fetch_recent_runs_per_env(environment_id, starting_from)

        for run in runs:
            run_id = run["id"]
            _results = self._run_results_by_run(
                run_id, test_names, column_names
            )
            run_results.extend(_results)
        return run_results

    def _fetch_all_run_results(self, starting_from: str) -> SerializedAsset:
        run_results: list[dict] = []

        for environment in self._fetch_environments():
            environment_id = environment["id"]
            _results = self._run_results_by_env(environment_id, starting_from)
            run_results.extend(_results)

        return run_results

    def fetch(
        self, asset: CoalesceAsset, starting_from=None
    ) -> SerializedAsset:
        """Extract the given Coalesce Asset"""
        if asset in (CoalesceAsset.NODES, CoalesceQualityAsset.NODES):
            return self._fetch_all_nodes()
        elif asset == CoalesceQualityAsset.RUN_RESULTS:
            return self._fetch_all_run_results(starting_from=starting_from)
        raise AssertionError(
            f"Asset {asset} is not supported by CoalesceClient"
        )



from ....utils import SerializedAsset
from .type import NodeIDToNamesMapping

_NULL_SUFFIX = ": Null"
_UNIQUE_SUFFIX = ": Unique"


def is_test(
    query_result: dict,
    node_id: str,
    test_names: NodeIDToNamesMapping,
    column_names: NodeIDToNamesMapping,
) -> bool:
    """
    checks whether a query result is a test result or not.

    all this implementation can soon be replaced by checking whether
    query_result['type'] == 'sqlTest', which should be GA Apr 28th 2025
    """
    # test scoped on the node (table)
    result_name = query_result["name"]
    if result_name in test_names.get(node_id, {}):
        return True

    # test scoped on the column
    if result_name.endswith(_NULL_SUFFIX) or result_name.endswith(
        _UNIQUE_SUFFIX
    ):
        column_name = result_name.split(":")[0]
        if column_name in column_names.get(node_id, {}):
            return True
    return False


def test_names_per_node(nodes: SerializedAsset) -> NodeIDToNamesMapping:
    """mapping nodeID: set(testName)"""
    mapping: dict[str, set[str]] = {}
    for node in nodes:
        node_id = node["id"]
        tests = node.get("metadata", {}).get("appliedNodeTests", [])
        mapping[node_id] = {test["name"] for test in tests}
    return mapping


def column_names_per_node(nodes: SerializedAsset) -> NodeIDToNamesMapping:
    """mapping nodeID: set(columnNames)"""
    mapping: dict[str, set[str]] = {}
    for node in nodes:
        node_id = node["id"]
        columns = node.get("metadata", {}).get("columns", [])
        mapping[node_id] = {column["name"] for column in columns}
    return mapping



from .asset import (
    ADDITIONAL_LINEAGE_ASSETS,
    CATALOG_ASSETS,
    EXTERNAL_LINEAGE_ASSETS,
    FUNCTIONS_ASSETS,
    QUERIES_ASSETS,
    VIEWS_ASSETS,
    SupportedAssets,
    WarehouseAsset,
    WarehouseAssetGroup,
    extractable_asset_groups,
)
from .extract import SQLExtractionProcessor, common_args
from .query import (
    QUERIES_DIR,
    AbstractQueryBuilder,
    ExtractionQuery,
    TimeFilter,
)
